# Copyright (C) 2010 Thomson Reuters
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from bzrlib import ui, trace
from bzrlib.branch import Branch, BranchFormat, GenericInterBranch, PullResult
from bzrlib.bzrdir import BzrDirFormat, BzrDir
from bzrlib.errors import NoWorkingTree, NotLocalUrl, NotBranchError, UnstackableBranchFormat
from bzrlib.foreign import ForeignBranch
from bzrlib.repository import InterRepository
from bzrlib.revision import NULL_REVISION
from bzrlib.plugins.tfs.repository import TfsRepository
from bzrlib.plugins.tfs.mapping import BzrTfsMapping
from bzrlib.plugins.tfs.foreign import get_rich_root_format
from bzrlib.plugins.tfs.transport import get_tfs_transport

class TfsBranchFormat(BranchFormat):
    def __init__(self):
        BranchFormat.__init__(self)

    def network_name(self):
        return "tfs"

    def __get_matchingbzrdir(self):
        from bzrlib.plugins.tfs.foreign import TfsControlFormat
        return TfsControlFormat()

    _matchingbzrdir = property(__get_matchingbzrdir)

    def get_format_description(self):
        return "Microsoft TFS Server"

    def get_format_string(self):
        return "Microsoft TFS Server"

    def supports_tags(self):
        return False

    def get_foreign_tests_branch_factory(self):
        from bzrlib.plugins.tfs.tests.util import MockTfsBranchFactory
        return MockTfsBranchFactory()

    def get_foreign_tests_repository_factory(self):
        from bzrlib.plugins.tfs.tests.util import MockTfsRepositoryFactory
        return MockTfsRepositoryFactory()


class TfsBranch(ForeignBranch):
    def __init__(self, repository, branch_path, revnum=None, _skip_check=False, mapping=None):
        """Instantiate a new TfsBranch"""
        self.repository = repository
        self._format = TfsBranchFormat()
        super(TfsBranch, self).__init__(mapping or BzrTfsMapping())
        self._revnum = revnum
        self._branch_path = branch_path
        self._lock_mode = None
        self._lock_count = 0
        self._revmeta_cache = None
        self.base = self.repository.base

    def is_locked(self):
        return self._lock_count != 0
    def break_lock(self):
        pass
    def lock_write(self):
        if self._lock_mode:
            self._lock_count +=1
        else:
            self._lock_mode = 'w'
            self._lock_count = 1
        return self.repository.lock_write()
    def lock_read(self):
        if self._lock_mode:
            self._lock_count +=1
        else:
            self._lock_mode = 'r'
            self._lock_count = 1
        return self.repository.lock_read()
    def unlock(self):
        self._lock_count -=1
        if self._lock_count == 0:
            self._lock_mode = None
        self.repository.unlock()

    def get_physical_lock_status(self):
        """See Repository.get_physical_lock_status()."""
        return False

    def _get_nick(self, local=False, possible_master_transports=None):
        """Find the nick name for this branch.

        :return: Branch nick
        """
        bp = self._branch_path.strip("/")
        if self._branch_path == "":
            return self.base.split("/")[-1]
        return bp

    nick = property(_get_nick)

    def _gen_revision_history(self):
        #TODO: can we use itertools here?
        history = [rev.revision_id() for rev in self._revision_meta_history()]
        history.reverse()
        return history

    def _revision_meta_history(self):
        #svn approach is as follows:
        # create a cache of RevisionMetadata objects
        # each object converts between svn props and bzr props
        #
        if self._revmeta_cache is None:
            #self._revmeta_cache = util.lazy_readonly_list(
            self._revmeta_cache = list(
                    self.repository._fetch_tfs_history(
                        self._branch_path,
                        self._revnum or self.repository.get_latest_revnum(),
                        to_revnum = 0))
        return self._revmeta_cache

    def get_parent(self):
        "Getting parent"
        return None

    def set_parent(self, url):
        """See Branch.set_parent()."""

    def create_checkout(self, to_location, revision_id=None, lightweight=False,
                        accelerator_tree=None, hardlink=False):
        checkout_branch = BzrDir.create_branch_convenience(
            to_location, force_new_tree=False, format=get_rich_root_format())
        checkout = checkout_branch.bzrdir
        checkout_branch.bind(self)
        # pull up to the specified revision_id to set the initial
        # branch tip correctly, and seed it with history.
        checkout_branch.pull(self, stop_revision=revision_id)
        return checkout.create_workingtree(revision_id, hardlink=hardlink)

    def set_last_revision_info(self, revno, revision_id):
        "This is really BZR-specific, we don't do anything with it"
        pass

    def set_push_location(self, location):
        """See Branch.set_push_location()."""
        trace.mutter("setting push location for %s to %s", self.base, location)

    def get_push_location(self):
        """See Branch.get_push_location()."""
        # get_push_location not supported on Subversion
        return None

    def get_stacked_on_url(self):
        raise UnstackableBranchFormat(self._format, self.base)

class TfsDir(BzrDir):
    def __init__(self, _transport, _format=None):
        if _format is None:
            from bzrlib.plugins.tfs.foreign import TfsControlFormat
            _format = TfsControlFormat()

        self._format = _format

        #convert from a BZR transport to a TFS one, if needed
        _transport = get_tfs_transport(_transport)
        self.transport = None
        self.root_transport = _transport
        self.branch_path = self.root_transport.project

    def open_repository(self, _unsupported=False):
        return TfsRepository(self, self.root_transport)

    def open_workingtree(self, _unsupported=False, recommend_upgrade=True):
        """
        Bzr working trees never exist on TFS repositories
        """
        raise NoWorkingTree(self.root_transport.base)

    def create_workingtree(self, revision_id=None,hardlink=None):
        """
        Bzr working trees never exist on TFS repositories
        """
        raise NotLocalUrl(self.root_transport.base)

    def open_branch(self, unsupported=True, ignore_fallbacks=False, name=""):
        repos = self.find_repository()
        branch = TfsBranch(repos, self.branch_path)
        branch.bzrdir = self
        return branch

    def clone(self, url, revision_id=None, force_new_repo=False):
        raise NotImplementedError(TfsDir.clone)

    def cloning_metadir(self, require_stacking=False):
        return get_rich_root_format(require_stacking)

class TfsPullResult(PullResult):
    """TFS source pull result.

    The main difference with a standard Bazaar PullResult is that it also
    reports the TFS changeset number.
    """
    def report(self, to_file):
        if not trace.is_quiet():
            if self.old_revid in (self.new_revid, NULL_REVISION):
                to_file.write('No revisions to pull.\n')
            else:
                mapping = BzrTfsMapping()
                #actually we need to consult repository since might be bzr-native rev_id
                try:
                    _, changeset = mapping.revision_id_bzr_to_foreign(self.new_revid)
                except:
                    changeset = 'unknown'
                to_file.write('Now on revision %d (TFS changeset: %s).\n' %
                        (self.new_revno, changeset))

class FromTfsBranch(GenericInterBranch):
    """Implementation for copying revisions from TFS; used during pull"""

    @classmethod
    def is_compatible(self, source, target):
        #source is TFS, target is other Bzr format
        return (isinstance(source, TfsBranch) and not isinstance(target, TfsBranch))

    def update_revisions(self, stop_revision=None, overwrite=False, graph=None):
        # what's the last revision in the source we are pulling from?
        other_last_revision = self.source.last_revision()
        if stop_revision is None:
            stop_revision = other_last_revision

        if stop_revision == NULL_REVISION:
            #no commits, so we're done.
            return self.target.last_revision_info()

        # current last revision we know about
        last_rev = self.target.last_revision()
        pb = ui.ui_factory.nested_progress_bar()
        # fetch the revisions we don't have [needed=missing]
        try:
            inter = InterRepository.get(self.source.repository, self.target.repository)
            #TODO: figure out why fetch bails early on NULL_REVISION
            if last_rev == NULL_REVISION:
                inter.fetch(revision_id=None, pb=pb)
            else:
                inter.fetch(revision_id=last_rev, pb=pb)
        finally:
            pb.finished()
        #if we're not overwriting we need to check for descendants
        if not overwrite:
            if graph is None:
                graph = self.target.repository.get_graph()
            self.target.lock_read()
            try:
                if self.target._check_if_descendant_or_diverged(
                        stop_revision, last_rev, graph, self.source):
                    # stop_revision is a descendant of last_rev, but we
                    # aren't overwriting, so we're done.
                    return self.target.last_revision_info()
            finally:
                self.target.unlock()
        self.target.generate_revision_history(stop_revision)
        return self.target.last_revision_info()

    def pull(self, overwrite=False, stop_revision=None, possible_transports = None, local=False):
        result = TfsPullResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.old_revno, result.old_revid = self.target.last_revision_info()
        if stop_revision == NULL_REVISION:
            result.new_revid = NULL_REVISION
        #walk backwards to find best revision
        #elif not stop_revision:
        #    stop_revision = result.old_revid

        #delegated to _update_revisions
        (result.new_revno, result.new_revid) = self.update_revisions(stop_revision, overwrite)
        #run any post_pull hooks
        return result
