# Copyright (C) 2010 Thomson Reuters
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from bzrlib.repository import InterRepository
from bzrlib.plugins.tfs.repository import TfsRepository
from bzrlib.plugins.tfs.mapping import BzrTfsMapping
from bzrlib.revision import NULL_REVISION
from bzrlib.foreign import ForeignRevision
from bzrlib.inventory import Inventory, InventoryDirectory, InventoryFile, InventoryLink
from bzrlib.versionedfile import FulltextContentFactory
from bzrlib import osutils, errors
import time
import os

class FromTfsRepository(InterRepository):
    @staticmethod
    def is_compatible(source, target):
        return isinstance(source, TfsRepository) and target.supports_rich_root()

    @staticmethod
    def _get_repo_format_to_test():
        return None

    def _determine_parent_id(self, parent_path, base_inv, new_dirs, root_id):
        #check dirs created in this delta
        #we check new dirs first to make sure
        #this allows us to parent to renamed dirs correctly
        if parent_path:
            if parent_path in new_dirs:
                return new_dirs[parent_path]

        #look up the parent id if it already exists
        if base_inv is not None and base_inv.has_filename(parent_path):
            return base_inv.path2id(parent_path)

        if parent_path:
            #we have a problem (possibly)
            #fall through to root for now
            #TODO: this happens with a Rename+Delete
            print "ERROR - missing parent", parent_path

        #assume the parent is the root
        return root_id

    def _path_relative_to_branch(self, path, branch):
        #is item part of path?
        if not path.startswith(branch):
            return None
        #trim off the location prefix
        path = path[len(branch):]
        #trim off a leading slash if present
        if path.startswith('/'):
            path = path[1:]
        return path

    def _delta_for_delete(self, actions, path, file_id, base_inventory):
        #handle Rename + Delete - we need to delete from the old location
        if "Rename" in actions:
            if base_inventory is not None and file_id in base_inventory:
                path = base_inventory.id2path(file_id)

        #Don't bother deleting a file that isn't currently available
        if base_inventory is not None and file_id not in base_inventory:
            return None

        # standard delete - just drop the file
        return (path, None, file_id, None)
        #handle recursive deletions
        #if item_type == "Folder":
        #    children = base_inventory[file_id].children.values()
        #    while children:
        #        ie = children.pop()
                #recursively delete if not explicitly actioned
        #        if ie.file_id not in used_file_ids:
        #            del_op = (base_inventory.id2path(ie.file_id), None, ie.file_id, None)
        #            delta.append(del_op)
        #            used_file_ids[ie.file_id] = (["recursive-delete"], path, del_op)
        #            if ie.kind=='directory':
        #                children.extend(ie.children.values())

    def build_inventory_for_revision(self, revision, base_inventory, parents):
        revid = revision.revision_id()
        branch = self.source.transport.project

        #create an inventory (to be used if no base inventory
        inventory = Inventory(revision_id = revid)
        delta = []
        new_dirs = {}
        used_file_ids = {}
        recursively_deleted = []

        #fetch the TFS delta
        changes = self.source.get_delta(revision.changeset)
        #print "Delta", revision.changeset
        #print changes #for debugging, easier than raw XML on my eyes

        #actions = None,Add,Edit,Encoding,Rename,Delete,Undelete,Branch,Merge,Lock
        # Add,Edit,Rename,Delete handled
        # Undelete - handle as an Add (file re-added to tree)
        # Encoding - ignored on purpose
        # None (what causes this?) - ignored for now
        # Merge, Lock  - ignored for now
        for (actions, path, item_type, item_id, artifact_url) in changes:

            path = self._path_relative_to_branch(path, branch)
            if not path:
                continue

            parent_path, name = os.path.split(path)
            parent_id = self._determine_parent_id(parent_path, base_inventory, new_dirs, inventory.root.file_id)

            if base_inventory is not None and base_inventory.has_filename(path):
                # lookup the file id if it already exists
                file_id = base_inventory.path2id(path)
                # make sure we generate a new file_id if this is an add!
                #   this can happen if an exisitng file is renamed
                #   and a new file is added with the old name
                if "Add" in actions:
                    file_id = "TFS-" + str(item_id)
            else:
                #else assign an ID ourselves corresponding to the TFS id
                file_id = "TFS-" + str(item_id)

            if file_id in used_file_ids:
                existing_op = used_file_ids[file_id]
                #remove an actioned recursive delete if we happen to be doing a rename
                if "Delete" not in actions and "recursive-delete" in existing_op[0]:
                    delta.remove(existing_op[2])
                #if we've already recursively deleted this file, we can skip the rest
                elif "Delete" in actions and "recursive-delete" in existing_op[0]:
                    continue
                else:
                    raise errors.BzrError("CONFLICT:\n%s\n%s, %s, %s" % (existing_op, actions, path, item_id))
            else:
                used_file_ids[file_id] = (actions, path, item_id)

            # delete, so we don't have to check contents
            if "Delete" in actions:
                op = self._delta_for_delete(actions, path, file_id, base_inventory)
                if op:
                    delta.append(op)
                    #handle recursive deletions
                    if item_type == "Folder":
                        children = base_inventory[file_id].children.values()
                        while children:
                            ie = children.pop()
                            #recursively delete if not explicitly actioned
                            if ie.file_id not in used_file_ids:
                                del_op = (base_inventory.id2path(ie.file_id), None, ie.file_id, None)
                                delta.append(del_op)
                                used_file_ids[ie.file_id] = (["recursive-delete"], path, del_op)
                                if ie.kind=='directory':
                                    children.extend(ie.children.values())
                continue

            if item_type == "Folder":
                #inventory entry for a folder
                ie = InventoryDirectory(file_id, name, parent_id)
                self.target.texts.add_lines((file_id, revid), [], [])
            else:
                #inventory entry for a file
                ie = InventoryFile(file_id, name, parent_id)
                contents = self.source.get_file_contents(artifact_url)
                record = FulltextContentFactory((file_id, revid), [], osutils.sha_string(contents), contents)
                ie.text_sha1 = record.sha1
                ie.text_size = len(contents)
                self.target.texts.insert_record_stream([record])

            ie.revision = revid

            # "Branch" may merely be a copy from another part of the tree
            # For now treating as an add but need to look at preserving history
            if "Add" in actions or "Branch" in actions or "Undelete" in actions:
                delta.append((None, path, file_id, ie))
                if item_type == "Folder":
                    new_dirs[path] = file_id
            elif "Rename" in actions:
                #avoid double rename from TFS (TODO: how does this happen??)
                existing_path = [d for d in delta if d[1] == path]
                if existing_path:
                    continue
                if base_inventory is not None and file_id in base_inventory:
                    old_path = base_inventory.id2path(file_id)
                else:
                    #rename from another part of the tree
                    #therefore we don't have access to old version, treat as an add
                    old_path = None
                if item_type == "Folder":
                    new_dirs[path] = file_id
                delta.append((old_path, path, file_id, ie))
            elif "Edit" in actions:
                delta.append((path, path, file_id, ie))

        #use the delta to generate a new inventory
        if base_inventory:
            base_rev = parents[0]
            (validator, new_inventory) = self.target.add_inventory_by_delta(base_rev, delta, revid, (), base_inventory)
            return new_inventory

        #no old inventory, so create a new one
        inventory.root.revision = revid
        #don't forget the text for a root entry!
        self.target.texts.add_lines((inventory.root.file_id, revid), [], [])
        inventory.apply_delta(delta)
        self.target.add_inventory(revid, inventory, [])
        return inventory

    def fetch(self, revision_id =None, pb = None, find_ghosts = False, 
            needed = None, mapping = None, project = None, fetch_spec=None):
        #TODO: figure out why we are bailing early
        if revision_id == NULL_REVISION:
            return

        if pb:
            pb.update("fetch phase", 0, 2)

        self.target.lock_write()
        try:
            if needed is None:
                #get list of revisions that are in self.source, but not in self.target
                #if revision_id, find until
                #if fetch_spec, for fs.head: find until
                #else just get everything
                latest = self.source.get_latest_revnum()

                #TODO: iter over source history until we find a TFS id
                # use that rather than trying to fetch full history!
                revs = self.source._fetch_tfs_history(None, latest, 0)
                candidates = [(rev, rev.revision_id()) for rev in revs]
                already_fetched = self.target.has_revisions([c[1] for c in candidates])
                needed = [c[0] for c in candidates if c[1] not in already_fetched]

            if len(needed) == 0:
                #nothing to fetch
                return
            if pb:
                pb.update("fetch phase", 1, 2)

            needed.reverse()
            last_rev = revision_id
            last_inv = None
            for (i, trev) in enumerate(needed):
                if pb is not None:
                    pb.update("Fetching revision", i)
                revid = trev.revision_id()
                #if revision_id:
                #    print "Need", trev.changeset
                if last_rev:
                    parent = (last_rev,)
                    #on an update this can happen on the first pass
                    if not last_inv:
                        last_inv = self.target.get_inventory(last_rev)
                        assert last_inv
                else:
                    parent = ()
                    last_inv = None
                last_rev = revid
                #get the rev info
                rev = ForeignRevision(foreign_revid = trev.changeset, mapping = BzrTfsMapping(), revision_id = revid, parent_ids = parent)
                #set changed file props
                #set revision properties
                rev.committer = trev.author
                rev.message = trev.comment
                ts, tz = trev.bzr_time()
                rev.timestamp = ts
                rev.timezone = tz
                self.target.start_write_group()
                try:
                    inventory = self.build_inventory_for_revision(trev, last_inv, parent)
                    #rev.inventory_sha1 = "" #hmmm...how do we calculate this?
                    self.target.add_revision(revid, rev, inventory)
                except:
                    self.target.abort_write_group()
                    raise
                else:
                    self.target.commit_write_group()
                last_inv = self.target.get_inventory(revid)
        finally:
            self.target.unlock()

class ToTfsRepository(InterRepository):
    @staticmethod
    def is_compatible(source, target):
        return isinstance(target, TfsRepository)

    @staticmethod
    def _get_repo_format_to_test():
        return None

    def copy_content(self, revision_id = None, pb = None):
        todo = []
        for revision_id in self.source.iter_reverse_revision_history(revision_id):
            #If we break here we're assuming that any changes have been rebased
            if self.target.has_revision(revision_id):
                break
            todo.append(revision_id)
        print "pushing revs:", todo

        for revid in todo:
            self.create_changeset_from_revision(revid)
        #TODO: look at using source.get_deltas_for_revisions(todo)
        # guessing this is what dpush is using?
        # look @ source for push, dpush commands
        return

    def pull(self, overwrite=False, stop_revision=None, possible_transports = None, local=False):
        print "Pulling until", stop_revision

    def fetch(self, revision_id =None, pb = None, find_ghosts = False, fetch_spec=None):
        if fetch_spec is not None:
            print "UNEXPECTED:", fetch_spec
        self.copy_content(revision_id = revision_id, pb=pb)

    def create_changeset_from_revision(self, revision_id):
        #generate a delta
        print "getting delta"
        delta = self.source.get_revision_delta(revision_id)
        rev = self.source.get_revision(revision_id)

        #TODO: do we need to sort by path, or will server do that?
        pendops = self.delta_to_pending_operations(delta)
        #print pendops
        texts = self.lookup_changed_texts(revision_id, delta)

        #create a label with bzr rev_id
        self.target.transport.commit_delta(pendops, texts, rev)

    def delta_to_pending_operations(self, delta):
        pendops = []
        #push to server
        for change in delta.added:
            pendops.append((self.target.transport.project + "/" + change[0], "Add", change[2]))
        for change in delta.removed:
            pendops.append((self.target.transport.project + "/" + change[0], "Delete", change[2]))
        for change in delta.renamed:
            #TODO: handle pushing renames!
            print "Rename", change
        for change in delta.modified:
            pendops.append((self.target.transport.project + "/" + change[0], "Edit", change[2]))
        return pendops

    # for added and modified items, we need the byte contents to upload
    def lookup_changed_texts(self, revision_id, delta):
        inv = self.source.get_inventory(revision_id)
        texts = {}
        desired = []
        for path, file_id, type in delta.added:
            if type == "file":
                ie = inv[file_id]
                desired.append((file_id, ie.revision, self.target.transport.project + "/" + path))
        for path, file_id, type, _, _ in delta.modified:
            if type == "file":
                ie = inv[file_id]
                desired.append((file_id, ie.revision, self.target.transport.project + "/" + path))
        for item, bytes in self.source.iter_files_bytes(desired):
            texts[item] = bytes
        return texts

