# Copyright (C) 2010 Thomson Reuters
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from bzrlib.foreign import ForeignVcs, VcsMappingRegistry
from bzrlib import errors as bzr_errors
from bzrlib.repository import RepositoryFormat

class TfsRepositoryFormat(RepositoryFormat):
    rich_root_data = True
    supports_tree_reference = False
    _serializer = None

    def __init__(self):
        super(TfsRepositoryFormat, self).__init__()

    def get_format_description(self):
        return "TFS Repository"

    def network_name(self):
        return "tfs"

    def initialize(self, url, shared=False, _internal=False):
        raise bzr_errors.UninitializableFormat(self)

    def check_conversion_target(self, target_repo_format):
        if not target_repo_format.rich_root_data:
            raise bzr_errors.BadConversionTarget('Does not support rich root data.', target_repo_format)

    def get_foreign_tests_repository_factory(self):
        from bzrlib.plugins.tfs.tests.util import MockTfsRepositoryFactory
        return MockTfsRepositoryFactory()

from bzrlib.bzrdir import BzrDirFormat, format_registry
from bzrlib.lockable_files import TransportLock
from bzrlib.transport.local import LocalTransport

class TfsControlFormat(BzrDirFormat):
    _lock_class = TransportLock

    def __init__(self):
        super(TfsControlFormat, self).__init__()
        self.repository_format = TfsRepositoryFormat()

    def is_supported(self):
        return True

    def get_format_string(self):
        return 'Team Foundation Server'

    def get_format_description(self):
        return 'Team Foundation Server'

    def _open(self, transport):
        #TODO: create the actual directory representation
        from bzrlib.plugins.tfs.branch import TfsDir
        return TfsDir(transport, self)

    @classmethod
    def probe_transport(klass, transport):
        format = klass()
        if isinstance(transport, LocalTransport):
            #see how this goes?
            raise bzr_errors.NotBranchError(path=transport.base)
        if not transport.base.startswith("tfs+"):
            raise bzr_errors.NotBranchError(path=transport.base)
        return format

class TfsMappingRegistry(VcsMappingRegistry):
    def revision_id_bzr_to_foreign(self, revid):
        mapping = self.get('v1')()
        path, changeset = mapping.revision_id_bzr_to_foreign(revid)
        return (changeset, mapping)

mapping_registry = TfsMappingRegistry()
mapping_registry.register_lazy('v1', 'bzrlib.plugins.tfs.mapping', 'BzrTfsMapping', 'Original Bazaar-TFS mapping')
mapping_registry.set_default('v1')

class ForeignTfs(ForeignVcs):

    def __init__(self):
        super(ForeignTfs, self).__init__(mapping_registry)
        self.abbreviation = "tfs"

    @property
    def branch_format(self):
        from bzrlib.plugins.tfs.branch import TfsBranchFormat
        return TfsBranchFormat()

    @property
    def repository_format(self):
        return TfsRepositoryFormat()

    #TODO: should show server URL in v2 mapping?
    def show_foreign_revid(self, foreign_id):
        return {"TFS changeset": foreign_id}

foreign_vcs_tfs = ForeignTfs()

def get_rich_root_format(stacked=False):
    format = format_registry.make_bzrdir('default-rich-root')
    #TODO: need to handle old clients
    return format
