# Copyright (C) 2010 Thomson Reuters
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import urllib
import datetime, calendar
import re
import sys

from bzrlib import transactions
from bzrlib.foreign import ForeignRepository
from bzrlib.repository import Repository
from bzrlib.plugins.tfs.foreign import foreign_vcs_tfs, TfsRepositoryFormat
from bzrlib.lock import _RelockDebugMixin, LogicalLockResult

#this class copied 
class DummyLockableFiles(object):
    def __init__(self, transport):
        self._transport = transport
    def create_lock(self):
        pass
    def break_lock(self):
        pass
    def leave_in_place(self):
        pass
    def dont_leave_in_place(self):
        pass
    def lock_write(self, token=None):
        pass
    def lock_read(self):
        pass
    def unlock(self):
        pass
    def is_locked(self):
        return False
    def get_physical_lock_status(self):
        return False
    def get_transaction(self):
        return transactions.PassThroughTransaction()


def parse_timestamp(s):
    """Returns (datetime, tz offset in minutes) or (None, None)."""
    m = re.match(""" ^
        (?P<year>-?[0-9]{4}) - (?P<month>[0-9]{2}) - (?P<day>[0-9]{2})
        T (?P<hour>[0-9]{2}) : (?P<minute>[0-9]{2}) : (?P<second>[0-9]{2})
        (?P<microsecond>\.[0-9]{1,6})?
        (?P<tz>
        Z | (?P<tz_hr>[-+][0-9]{2}) : (?P<tz_min>[0-9]{2})
        )?
        $ """, s, re.X)
    if m is not None:
        values = m.groupdict()
    if values["tz"] in ("Z", None):
        tz = 0
    else:
        tz = int(values["tz_hr"]) * 60 + int(values["tz_min"])
    if values["microsecond"] is None:
        values["microsecond"] = 0
    else:
        values["microsecond"] = values["microsecond"][1:]
        values["microsecond"] += "0" * (6 - len(values["microsecond"]))
    values = dict((k, int(v)) for k, v in values.iteritems()
                  if not k.startswith("tz"))
    try:
        return datetime.datetime(**values), tz
    except ValueError:
        pass
    return None, None


class TfsRevision:
    def __init__(self, changeset, path, author, timestamp, comment):
        self.changeset = int(changeset)
        self.path = path
        self.author = author
        self.timestamp = timestamp
        if comment and "\n\nbzr-revid:" in comment:
            self.comment, self.bzrid = comment.split("\n\nbzr-revid:")
        else:
            self.comment = comment or ''
            self.bzrid = None
    def revision_id(self):
        # if a bzr revid was found, use that
        if self.bzrid:
            return self.bzrid
        # otherwise construct a revision ID
        #TODO: use mapping class
        return "tfs-v1:%s:%d" % (urllib.quote(self.path), self.changeset)
    def bzr_time(self):
        ts, tz = parse_timestamp(self.timestamp)
        ts = calendar.timegm(ts.timetuple()) #convert to UNIX time
        tz = tz *60 #convert minutes to seconds
        return ts, tz

class TfsRepository(ForeignRepository):
    def __init__(self, bzrdir, transport, branch_path=None):
        self.vcs = foreign_vcs_tfs
        self._revision_store = []
        self._revstore_min = sys.maxint
        self._revstore_max = -1
        self._bzr_ids = {}
        control_files = DummyLockableFiles(transport)
        Repository.__init__(self, TfsRepositoryFormat(), bzrdir, control_files)
        self._lock_mode = None
        self._lock_count = 0
        self.transport = transport
        self._serializer = None
        #self.uuid = transport.get_uuid()
        if hasattr(transport, 'bzr_url'):
            self.base = transport.bzr_url

    def is_locked(self):
        return self._lock_count != 0
    def is_write_locked(self):
        return self._lock_mode == 'w'
    def break_lock(self):
        pass
    def lock_write(self):
        if self._lock_mode:
            self._lock_count +=1
        else:
            self._lock_mode = 'w'
            self._lock_count = 1
        return LogicalLockResult(self.unlock)
    def lock_read(self):
        if self._lock_mode:
            self._lock_count +=1
        else:
            self._lock_mode = 'r'
            self._lock_count = 1
        return LogicalLockResult(self.unlock)
    def unlock(self):
        self._lock_count -=1
        if self._lock_count == 0:
            self._lock_mode = None

    def is_shared(self):
        """Return True if this repository is flagged as a shared repository."""
        return True

    def get_physical_lock_status(self):
        """See Repository.get_physical_lock_status()."""
        return False

    def gather_stats(self, revid=None, committers=None):
        """See Repository.gather_stats()."""
        result = {}
        # Approximate number of revisions
        result['revisions'] = self.get_latest_revnum()+1
        return result

    def make_working_trees(self):
        #Working trees are never created inside TFS repositories.
        return False

    def _fetch_tfs_history(self, branch_path, revnum, to_revnum):
        if not branch_path:
            branch_path = self.transport.project
        history = self.transport.get_log(branch_path, revnum, to_revnum)
        cache_info = [TfsRevision(changeset.revision, branch_path, changeset.author, changeset.time, changeset.comment) for changeset in history]
        self._revision_store.extend(cache_info)
        self._revstore_max = max(self._revstore_max, revnum)
        if (len(cache_info) > 0): # there could be no revisions
            self._revstore_max = max(self._revstore_max, cache_info[0].changeset)
            self._revstore_min = min(cache_info[-1].changeset, self._revstore_min)
        #print self._revstore_min, self._revstore_max
        self._bzr_ids.update(dict([(r.bzrid, r.changeset) for r in cache_info if r.bzrid]))
        return cache_info

    def get_delta(self, rev):
        return self.transport.get_delta(rev)

    def get_file_contents(self, artifact_url):
        return self.transport.get_file_contents(artifact_url)


    def get_latest_revnum(self):
        return self.transport.get_latest_revnum()

    def has_revision(self, revision_id):
        #if TFS native easy enough...
        if revision_id.startswith("tfs-"):
            return True
        #deprecated - old experimental (0.1) prefix
        if revision_id.startswith("tfs:"):
            return True
        #TODO: actually need to query history and
        #inspect revids
        if revision_id in self._bzr_ids:
            return True
        return False

