# Copyright (C) 2010 Thomson Reuters
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from bzrlib.transport import Transport
from bzrlib.errors import NoSuchFile, TransportNotPossible
from bzrlib.plugins.tfs.vcsclient import TfsRepositoryWrapper


def get_tfs_transport(bzr_transport):
    """Given a standard Bazaar transport, create a corresponding TFS transport"""

    # actually a TFS transport
    if isinstance(bzr_transport, TfsTransport):
        return bzr_transport

    #use cached value if available
    cached = getattr(bzr_transport, "_tfs", None)
    if cached is not None:
        return cached

    tfs_url = bzr_transport.external_url()
    #TODO: svn plugin is smart enough to reuse credentials from HTTP transport
    transport = TfsTransport(tfs_url)
    #cache for future use
    bzr_transport._tfs = transport
    return transport

# not a bzr transport; we don't want to be picked up by the standard
# unit tests
def get_test_permutations():
    return []

class TfsTransport(Transport):
    """Handle the TFS transport scheme to allow talking to remote servers"""
    def __init__(self, url, from_transport=None, credentials=None):
        self.bzr_url = url
        self.server_url, self.project = url.split('$')
        self.project = '$' + self.project
        if self.server_url.startswith("tfs+"):
            self.server_url = self.server_url[len("tfs+"):]
        Transport.__init__(self, url)
        self.capabilities = {}
        #TODO: move the wrapper capabilities into this class?
        self.wrapper = TfsRepositoryWrapper(self.server_url)

        #TODO: do the same kind of async connection pooling the SVN plugin does

    ### Override what we need for Bzr to be happy
    def external_url(self):
        return self.bzr_url

    def get(self, relpath):
        raise NoSuchFile(path=relpath)

    def has(self, relpath):
        print "transport.has"

    def stat(self, relpath):
        print "transport.stat"

    def put_file(self, name, file, mode=0):
        print 'transport.put'

    def get_repos_root(self):
        print 'get_repos_root'

    def clone(self, offset=None):
        return TfsTransport(self.bzr_url)

    def abspath(self, relpath):
        print 'transport.abspath'

    def get_log(self, path, from_rev, to_rev):
        if not path:
            path = self.project
        history = []
        total = from_rev - to_rev
        while total > 0:
            batch_size = min([total, 256])
            #print "getting", batch_size, "log entries", from_rev, "to", to_rev
            log = self.wrapper.get_log(path, from_rev, to_rev, batch_size)
            log_size = len(log)
            history.extend(log)
            if log_size < batch_size:
                break
            total -= log_size
            from_rev = int(log[-1].revision) - 1
        return history

    def get_delta(self, rev):
        return self.wrapper.get_delta(rev)

    def get_file_contents(self, artifact_url):
        return self.wrapper.get_file_contents(artifact_url)

    def get_latest_revnum(self):
        return self.wrapper.get_latest_version()

    def commit_delta(self, delta, texts, rev):
        wsname = rev.revision_id
        try:
            workspace = self.wrapper.create_workspace(wsname, self.project)
            self.wrapper.get_existing(wsname, delta)
            #pend adds
            self.wrapper.pend_changes(wsname, [d for d in delta if d[1] == "Add"])
            #pend edits
            self.wrapper.pend_changes(wsname, [d for d in delta if d[1] == "Edit"])
            #pend removes
            self.wrapper.pend_changes(wsname, [d for d in delta if d[1] == "Delete"])

            cset = self.wrapper.finish_commit(wsname, self.project, delta, rev, texts)
            #TODO: convert to trace calls
            if cset:
                print "Committed changeset", cset
            else:
                print "COMMIT FAILED"
        finally:
            self.wrapper.delete_workspace(wsname)
