# Copyright (C) 2010 Thomson Reuters
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import urllib2
import zlib, gzip, StringIO
import mimetools, hashlib, base64
import datetime
from urlparse import urlparse
try:
    from xml.etree import ElementTree as ET
    from xml.etree.ElementTree import Element, SubElement, QName
except: #Python 2.4
    from elementtree import ElementTree as ET
    from elementtree.ElementTree import Element, SubElement, QName

from bzrlib import config, errors
from ntlm import HTTPNtlmAuthHandler

# namespaces (SOAP 1.2)
NS_SOAP_ENV = "{http://www.w3.org/2003/05/soap-envelope}"
NS_XSI = "{http://www.w3.org/1999/XMLSchema-instance}"
NS_XSD = "{http://www.w3.org/1999/XMLSchema}"
# namespaces (TFS-specific)
NS_TFS = "{http://schemas.microsoft.com/TeamFoundation/2005/06/VersionControl/ClientServices/03}"
NS_TFS_REG = "{http://schemas.microsoft.com/TeamFoundation/2005/06/Services/Registration/03}"

def SoapRequest(method):
    # create a SOAP request element
    request = Element(method)
    request.set(
        NS_SOAP_ENV + "encodingStyle",
        "http://schemas.xmlsoap.org/soap/encoding/"
        )
    return request

def SoapElement(parent, name, type=None, text=None):
    # add a typed SOAP element to a request structure
    elem = SubElement(parent, name)
    if type:
        if not isinstance(type, QName):
            type = QName("http://www.w3.org/1999/XMLSchema", type)
        elem.set(NS_XSI + "type", type)
    elem.text = text
    return elem

class SoapService:
    def __init__(self, url=None):
        self.__client = HTTPClient(url or self.url)
    def call(self, action, request):
        # build SOAP envelope
        envelope = Element(NS_SOAP_ENV + "Envelope")
        body = SubElement(envelope, NS_SOAP_ENV + "Body")
        body.append(request)
        # call the server
        response = self.__client.do_request(
            ET.tostring(envelope),
            extra_headers=[("SOAPAction", action)]
            )
        return response.getroot().find(body.tag)[0]


class Message:
    def __init__(self, url, action, ns = NS_TFS):
        self.req = urllib2.Request(url)
        self.req.add_header("X-TFS-Version", "1.0.0.0")
        self.req.add_header("Content-Type", "application/soap+xml; charset=utf-8")
        self.req.add_header("accept-language", "en-US")
        #self.req.add_header("X-VersionControl-Instance", "ac4d8821-8927-4f07-9acf-adbf71119886")
        self.action = action
        self.body = Element(ns + action)

    def post(self, opener, dump=False):
        # build SOAP envelope
        envelope = Element(NS_SOAP_ENV + "Envelope")
        body = SubElement(envelope, NS_SOAP_ENV + "Body")
        body.append(self.body)

        ET._namespace_map[NS_TFS[1:-1]] = "tfs"
        ET._namespace_map[NS_XSI[1:-1]] = "xsi"
        self.req.add_data(ET.tostring(envelope))
        if dump:
            print ET.tostring(envelope)
            print "+"*20
        response = opener.open(self.req)
        if dump:
            print response.info()
        try:
            the_page = ET.parse(response)
        except:
            print response.read()
            raise
        return the_page.getroot()#.find(body.tag)[0]

class TfsRepositoryWrapper:
    """
    Handle the actual work of constructing appropriate SOAP messages to the TFS web service

    TODO:
      tag management - QueryLabels, LabelItem, DeleteLabel, UnlabelItem
    """
    def __init__(self, server_base):
        self.server_base = server_base
        if server_base[-1] != "/":
            server_base += "/"
        self.server_url = server_base + 'VersionControl/v1.0/Repository.asmx'
        self.registration_url = server_base + 'services/v1.0/Registration.asmx'
        self.download_url = server_base + 'VersionControl/v1.0/item.asmx'
        self.upload_url = server_base + 'VersionControl/v1.0/upload.asmx'
        self.version = None #will be set when we need to determine it
        self.ntlm()

    def ntlm(self):
        passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
        default_user = os.environ["USERNAME"]
        auth_conf = config.AuthenticationConfig()
        urlinfo = urlparse(self.server_base)
        #protocol, host, user, port, path, realm, prompt
        scheme = urlinfo[0]
        if ':' in urlinfo[1]:
            hostname, port = urlinfo[1].split(':')
            port = int(port)
        else:
            hostname = urlinfo[1]
            port = 80
            if scheme == "https":
                port = 443
        self.user = auth_conf.get_user(scheme, hostname, port = port, prompt="TFS username", default=default_user, ask = True)
        password = auth_conf.get_password(scheme, hostname, self.user, port = port, prompt = "%(user)s TFS password")
        passman.add_password(None, self.server_base, self.user, password)
        auth_NTLM = HTTPNtlmAuthHandler.HTTPNtlmAuthHandler(passman)
        opener = urllib2.build_opener(auth_NTLM)
        self.opener = opener

    def list_workspaces(self, user = None):
        msg = Message(self.server_url, "QueryWorkspaces")
        if not user: user = self.user
        if user:
            self.elem(msg.body, "ownerName", user)
        response = self.invoke(msg)
        workspaces = []
        for w in response.getiterator(NS_TFS + "Workspace"):
            computer, owner = w.attrib['computer'], w.attrib['owner']
            for wf in w.findall(".//" + NS_TFS + "WorkingFolder"):
                #TODO: use a named tuple or class
                workspaces.append((wf.attrib['item'], wf.attrib['local'], computer, owner))
        return workspaces

    def server_version(self):
        msg = Message(self.registration_url, "GetRegistrationEntries", ns=NS_TFS_REG)
        self.elem(msg.body, "toolId", "Framework")
        response = self.invoke(msg)
        entries = response.findall('.//' + NS_TFS_REG + 'RegistrationEntry')
        if len(entries) > 0:
            return "2010"

        # toolId = vstfs
        # if response. //ServiceInterface @name=GroupSecurity2, return
        # 2008
        return "2005"

    def server_info(self):
        msg = Message(self.server_url, "GetRepositoryProperties")
        response = self.invoke(msg)
        details = response.find('.//' + NS_TFS + "GetRepositoryPropertiesResult")
        guid = details.attrib['id']
        return guid

    def elem(self, parent, name, val):
        e = SubElement(parent, NS_TFS + name)
        e.text = str(val)

    def get_delta(self, revid):
        msg = Message(self.server_url, "QueryChangesForChangeset")
        self.elem(msg.body, "changesetId", revid)
        self.elem(msg.body, "generateDownloadUrls", str(True).lower())
        response = self.invoke(msg)
        changes = response.findall('.//' + NS_TFS + 'Change')
        delta = []
        for change in changes:
            actions = change.attrib['type'].split(' ')
            info = change.find(NS_TFS + "Item")
            path = info.attrib['item']
            item_type = info.attrib['type']
            item_id = info.attrib['itemid']
            if item_type == "File":
                durl = info.attrib['durl']
            else:
                durl = None
            deletion_id = int(info.get('did', 0))
            if deletion_id:
                if "Delete" not in actions:
                    continue
            delta.append((actions,path,item_type,item_id, durl))
        return delta

    def check_server_version(self):
        if self.version == None:
            self.version = self.server_version()
            if self.version == "2010":
                # these become .ashx calls
                self.download_url = self.download_url.replace('.asmx', '.ashx')
                self.upload_url = self.upload_url.replace('.asmx', '.ashx')

    def get_file_contents(self, artifact_url):
        self.check_server_version()
        url = "%s?%s" % (self.download_url, artifact_url)
        req = urllib2.Request(url)
        response = self.opener.open(req)
        data = response.read()
        if response.headers.get('Content-Type', '') == 'application/gzip':
            data = StringIO.StringIO(data)
            f = gzip.GzipFile(fileobj=data)
            data =f.read()
            f.close()
        response.close()
        return data

    def get_log(self, branch, from_rev, to_rev, max_count=10000):
        msg = Message(self.server_url, "QueryHistory")

        # add an itemSpec for the directory or file
        item = SubElement(msg.body, NS_TFS+"itemSpec")
        item.attrib["recurse"] = "Full"
        item.attrib["item"] = branch

        vi = SubElement(msg.body, NS_TFS + "versionItem")
        vi.attrib[NS_XSI + "type"] = "tfs:LatestVersionSpec"

        # bzr from_rev is newest, to_rev is oldest
        # other way round in TFS, so we switch the params
        if to_rev is not None:
            vf = SubElement(msg.body, NS_TFS + "versionFrom")
            vf.attrib[NS_XSI + "type"] = "tfs:ChangesetVersionSpec"
            vf.attrib["cs"] = str(to_rev)

        if from_rev:
            vto = SubElement(msg.body, NS_TFS + "versionTo")
            vto.attrib[NS_XSI + "type"] = "tfs:ChangesetVersionSpec"
            vto.attrib["cs"] = str(from_rev)

        self.elem(msg.body, "maxCount", max_count)
        self.elem(msg.body, "includeFiles", str(False).lower())
        self.elem(msg.body, "generateDownloadUrls", str(False).lower())
        self.elem(msg.body, "slotMode", str(False).lower())

        response = self.invoke(msg)
        log = response.findall('.//' + NS_TFS + 'Changeset')
        revisions = []
        for entry in log:
            rev = TfsRevisionMetadata(entry.attrib['cset'], entry.attrib['cmtr'], entry.attrib['date'], entry.find(NS_TFS + 'Comment').text)
            revisions.append(rev)
        return revisions


    def get_latest_version(self):
        msg = Message(self.server_url, "GetRepositoryProperties")
        response = self.invoke(msg)
        details = response.find('.//' + NS_TFS + "GetRepositoryPropertiesResult")
        return int(details.attrib['lcset'])

    def create_workspace(self, workspace, root_folder):
        msg = Message(self.server_url, "CreateWorkspace")
        ws = SubElement(msg.body, NS_TFS + 'workspace')
        ws.attrib['computer'] = 'bzr-bridge'
        ws.attrib['name'] = workspace
        ws.attrib['owner'] = self.user
        folders = SubElement(ws, NS_TFS + 'Folders')
        wf = SubElement(folders, NS_TFS + 'WorkingFolder')
        wf.attrib['local'] = '/bzr/wf'
        wf.attrib['item'] = root_folder
        response = self.invoke(msg)
        failures = response.findall('.//' + NS_TFS + 'Failure')
        if failures:
            err_msg = []
            for f in failures:
                err_msg.append(f.find(NS_TFS+'Message').text)
            raise errors.BzrError("\n".join(err_msg))

    def delete_workspace(self, workspaceName):
        msg = Message(self.server_url, "DeleteWorkspace")
        self.elem(msg.body, "workspaceName", workspaceName)
        self.elem(msg.body, "ownerName", self.user)
        response = self.invoke(msg)

    #pend changes
    def pend_changes(self, workspace, operations):
        if not operations:
            return
        msg = Message(self.server_url, "PendChanges")
        self.elem(msg.body, "workspaceName", workspace)
        self.elem(msg.body, "ownerName", self.user)
        c = SubElement(msg.body, NS_TFS + 'changes')
        for op in operations:
            path,req,type = op
            if type not in ["file", "folder"]:
                type = "file"
            #init cap
            type = type[0].upper() + type[1:]
            cr = SubElement(c, NS_TFS + 'ChangeRequest')
            cr.attrib["req"] = req
            cr.attrib["type"] = type
            cr.attrib["enc"] = "65001" # UTF-8, I think (used to be -2)
            item = SubElement(cr, NS_TFS + "item")
            item.attrib['item'] = path
        response = self.invoke(msg)
        failures = response.findall('.//' + NS_TFS + 'Failure')
        if failures:
            err_msg = []
            for f in failures:
                err_msg.append(f.find(NS_TFS+'Message').text)
            raise errors.BzrError("\n".join(err_msg))

    def get_existing(self, workspace, delta):
        msg = Message(self.server_url, "Get")
        self.elem(msg.body, "workspaceName", workspace)
        self.elem(msg.body, "ownerName", self.user)
        c = SubElement(msg.body, NS_TFS + 'requests')
        for path, action, type in delta:
            cr = SubElement(c, NS_TFS + 'GetRequest')
            item = SubElement(cr, NS_TFS + "ItemSpec")
            item.attrib['item'] = path
            item.attrib['recurse'] = 'None'
            ver = SubElement(cr, NS_TFS + "VersionSpec")
            #TODO: this would ideally be the changeset of the previous version
            ver.attrib[NS_XSI + "type"] = "tfs:LatestVersionSpec"
        #TODO: figure out what options are needed for CodePlex/2010 to be happy
        #self.elem(msg.body, 'force', "true")
        #self.elem(msg.body, 'options', "3")
        response = self.invoke(msg)
        ops = response.findall('.//' + NS_TFS + 'GetOperation')
        #itemid, tlocal, sver, titem
        local_info = []
        for o in ops:
            itemid = o.attrib['itemid']
            tlocal = o.attrib['tlocal']
            sver = o.attrib['sver'] #server version
            titem = o.attrib['titem']
            local_info.append((itemid,tlocal,sver,titem))
        if local_info:
            self.update_local_version(workspace, local_info)
        return local_info


    def update_local_version(self, workspace, ops):
        msg = Message(self.server_url, "UpdateLocalVersion")
        self.elem(msg.body, "workspaceName", workspace)
        self.elem(msg.body, "ownerName", self.user)
        c = SubElement(msg.body, NS_TFS + 'updates')
        for itemid, tlocal, ver, _ in ops:
            update = SubElement(c, NS_TFS + 'LocalVersionUpdate')
            update.attrib['itemid'] = itemid
            update.attrib['tlocal'] = tlocal
            update.attrib['lver'] = ver
        response = self.invoke(msg)

    def upload_file(self, workspace, path, data):
        data = "".join(data)
        # for added or edited files only
        self.check_server_version()
        req = urllib2.Request(self.upload_url)
        req.add_header("X-TFS-Version", "1.0.0.0")
        req.add_header("accept-language", "en-US")
        #req.add_header("X-VersionControl-Instance", "ac4d8821-8927-4f07-9acf-adbf71119886")
        boundary = mimetools.choose_boundary()
        req.add_header("Content-Type", "multipart/form-data; boundary=%s" % boundary)
        filesize=len(data)
        fields = {"item": path,
                "wsname":workspace,
                "wsowner": self.user,
                "filelength": filesize,
                "hash": base64.encodestring(hashlib.md5(data).digest()),
                #TODO: use range for streaming large files
                "range":"bytes=0-%d/%d" % (filesize, filesize)
                }
        body = self.encode_multipart(fields, data, boundary)
        req.add_data(body)
        response = self.opener.open(req)
        print path
        print response.info
        print "=" * 20


    def encode_multipart(self, fields, data, boundary):
        body = ""
        #encode the fields
        for k, v in fields.iteritems():
            body += '--%(boundary)s\r\nContent-Disposition: form-data; name="%(name)s"\r\n\r\n%(value)s\r\n' % {'boundary':boundary, 'name':k, 'value':v}
        #encode the file contents
        body += '--%(boundary)s\r\nContent-Disposition: form-data; name="content"; filename="item"\r\nContent-Type: application/octet-stream\r\n\r\n%(value)s\r\n' % {'boundary':boundary, 'value':data}
        #add the ending boundary
        body += '--%s--\r\n' % boundary
        return body


    def abort_commit(self):
        print "unpend changes"

    def finish_commit(self, workspace, root_path, delta, rev, texts):
        #upload any added/edited files
        for item, bytes in texts.iteritems():
            self.upload_file(workspace, item, bytes)

        msg = Message(self.server_url, "CheckIn")
        self.elem(msg.body, "workspaceName", workspace)
        self.elem(msg.body, "ownerName", self.user)
        c = SubElement(msg.body, NS_TFS + 'serverItems')

        for path, action, type in delta:
            self.elem(c, "string", path)

        info = SubElement(msg.body, NS_TFS + 'info')
        tzoffset = '%+03d:%02d' % (rev.timezone/3600, (rev.timezone/60) % 60)
        #info.attrib['date'] = datetime.datetime.fromtimestamp(rev.timestamp).replace(microsecond=0).isoformat() + tzoffset
        #info.attrib['owner'] = self.user
        info.attrib['cset'] = "0"
        self.elem(info, 'Comment', rev.message + "\r\n\r\nbzr-revid:" + rev.revision_id)
        cin = SubElement(info, NS_TFS + 'CheckinNote')
        v = SubElement(cin, NS_TFS + 'Values')
        coderev = SubElement(v, NS_TFS + 'CheckinNoteFieldValue')
        coderev.attrib['name'] = 'Code Reviewer'
        coderev.attrib['val'] = 'n/a'
        po = SubElement(info, NS_TFS + 'PolicyOverride')
        self.elem(info, "CheckinOptions", "ValidateCheckinOwner")

        response = self.invoke(msg, debug=True)
        print ET.tostring(response)
        result = response.find('.//' + NS_TFS + 'CheckInResult')
        if not result:
            print ET.tostring(response)
            return None
        cset = result.attrib['cset']
        failures = response.findall('.//' + NS_TFS + 'Failure')
        if not failures:
            return cset

        err_msg = []
        for f in failures:
            err_msg.append(f.find(NS_TFS+'Message').text)

        raise errors.BzrError("\n".join(err_msg))
        return None

    def list_labels(self, path):
        msg = Message(self.server_url, "QueryLabels")
        self.elem(msg.body, "filterItem", path)
        ver = SubElement(msg.body, NS_TFS + "versionFilterItem")
        ver.attrib[NS_XSI + "type"] = "tfs:LatestVersionSpec"
        self.elem(msg.body, "includeItems", "false")
        self.elem(msg.body, "generateDownloadUrls", "false")
        response = self.invoke(msg)
        labels = [l.get('name') for l in response.getiterator(NS_TFS + "VersionControlLabel")]
        return labels

    def set_label(self, workspace, path, label, revid):
        if not revid or revid=='0':
            return
        print "create label %s for cset %s" % (label, revid)
        msg = Message(self.server_url, "LabelItem")
        self.elem(msg.body, "workspaceName", workspace)
        self.elem(msg.body, "workpaceOwner", self.user)
        info = SubElement(msg.body, NS_TFS + 'label')
        info.attrib['date'] = datetime.datetime.utcnow().replace(microsecond=0).isoformat() + '+00:00'
        import urllib
        info.attrib['name'] = urllib.quote(label)
        info.attrib['lid'] = '0'
        ls = SubElement(msg.body, NS_TFS + 'labelSpecs')
        labelitemspec = SubElement(ls, NS_TFS+'LabelItemSpec')
        labelitemspec.attrib['ex'] = 'false'
        itemspec = SubElement(labelitemspec, NS_TFS+'ItemSpec')
        itemspec.attrib['item'] = path
        itemspec.attrib['recurse'] = 'None'
        ver = SubElement(labelitemspec, NS_TFS + "Version")
        ver.attrib[NS_XSI + "type"] = "tfs:ChangesetVersionSpec"
        ver.attrib["cs"] = str(revid)
        self.elem(msg.body, 'children', 'Replace')
        response = self.invoke(msg)
        failures = response.findall('.//' + NS_TFS + 'Failure')
        if not failures:
            return
        print 'FAILED LABEL'
        for f in failures:
            print f.find(NS_TFS+'Message').text

    def invoke(self, message, debug=False):
        #TODO: handle exceptions thrown during request or parse
        response = message.post(self.opener, debug)
        fault = response.find('.//' + NS_SOAP_ENV + 'Fault')
        if fault:
            faultstr = fault.find('faultstring')
            if not faultstr:
                faultstr = fault.find(NS_SOAP_ENV + 'Reason' + "/" + NS_SOAP_ENV + 'Text')
            if faultstr:
                msg = faultstr.text
            else:
                msg = ET.tostring(fault)
            raise errors.BzrError(msg)
        return response

class TfsRevisionMetadata:
    def __init__(self, revid, author, time, comment):
        self.revision = revid
        self.author = author
        self.time = time
        self.comment = comment
